@@ -1,6 +1,6 @@
-/* 
+/*
   Copyright(C) 2010 Tetsuro IKEDA
-  Copyright(C) 2010 Kentoku SHIBA
+  Copyright(C) 2010-2011 Kentoku SHIBA
   Copyright(C) 2011 Kouhei Sutou <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
@@ -38,6 +38,11 @@ extern "C" {
 #  define MRN_HANDLER_CLONE_NEED_NAME 1
 #endif
 
+#if (MYSQL_VERSION_ID >= 50603) || \
+    (MYSQL_VERSION_ID >= 50514)
+#  define MRN_HANDLER_HAVE_ADD_INDEX 1
+#endif
+
 #if MYSQL_VERSION_ID >= 50600
 #  define MRN_HANDLER_HAVE_HA_CLOSE 1
 #  define MRN_HANDLER_HAVE_HA_RND_NEXT 1
@@ -56,6 +61,16 @@ extern "C" {
   typedef Item COND;
 #endif
 
+#ifndef GRN_GEO_DEGREE2MSEC
+#  define GRN_GEO_DEGREE_ACCURACY 10000000
+#  define GRN_GEO_DEGREE2MSEC(degree)\
+    ((int)(((int)(degree) * 60 * 60 * 1000) +\
+           (((long)((degree) * GRN_GEO_DEGREE_ACCURACY) % GRN_GEO_DEGREE_ACCURACY) * 60 / GRN_GEO_DEGREE_ACCURACY * 60 * 1000) +\
+           (((long)((degree) * GRN_GEO_DEGREE_ACCURACY) % GRN_GEO_DEGREE_ACCURACY) * 60 % GRN_GEO_DEGREE_ACCURACY * 60 / 10000)))
+#  define GRN_GEO_MSEC2DEGREE(msec)\
+    ((((int)(msec)) / 3600.0) * 0.001)
+#endif
+
 class ha_mroonga;
 
 /* structs */
@@ -71,12 +86,13 @@ struct st_mrn_ft_info
   grn_ctx *ctx;
   grn_obj *table;
   grn_obj *result;
+  grn_obj *sorted_result;
   grn_obj *score_column;
   grn_obj key;
   grn_obj score;
+  uint active_index;
+  KEY *key_info;
   KEY *primary_key_info;
-  uint primary_key_length;
-  grn_id record_id; // TODO: remove me.
   ha_mroonga *mroonga;
 };
 
@@ -101,7 +117,8 @@ public:
   bool      is_clone;
   ha_mroonga *parent_for_clone;
   MEM_ROOT  *mem_root_for_clone;
-  grn_obj   pkey;
+  grn_obj   key_buffer;
+  grn_id    record_id;
 
 private:
   grn_ctx *ctx;
@@ -112,26 +129,20 @@ private:
   grn_obj **grn_index_tables;
   grn_obj **grn_index_columns;
 
-  grn_obj *result;
-  grn_obj *result0;
-  grn_table_cursor *cur;
-  grn_table_cursor *cur0;
-  grn_id record_id;
-  grn_obj *_score;
+  grn_obj *result_geo;
+  grn_table_cursor *cursor;
+  grn_table_cursor *index_table_cursor;
+  grn_obj *score_column;
+  grn_obj *key_accessor;
 
-  st_mrn_ft_info mrn_ft_info;
   grn_obj *matched_record_keys;
 
-  char **key_min;
-  char **key_max;
+  uchar **key_min;
+  uchar **key_max;
   int *key_min_len;
   int *key_max_len;
   uint dup_key;
 
-  longlong limit;
-  grn_table_sort_key *sort_keys;
-  int n_sort_keys;
-
   bool count_skip;
   bool fast_order_limit;
   bool fast_order_limit_with_index;
@@ -139,6 +150,12 @@ private:
   bool ignoring_duplicated_key;
   bool fulltext_searching;
 
+#ifdef MRN_HANDLER_HAVE_ADD_INDEX
+  handler_add_index *hnd_add_index;
+#endif
+
+  bool ignoring_no_key_columns;
+
 public:
   ha_mroonga(handlerton *hton, TABLE_SHARE *share);
   ~ha_mroonga();
@@ -270,6 +287,23 @@ public:
   int analyze(THD* thd, HA_CHECK_OPT* check_opt);
   int optimize(THD* thd, HA_CHECK_OPT* check_opt);
   bool is_fatal_error(int error_num, uint flags);
+  bool check_if_incompatible_data(HA_CREATE_INFO *create_info,
+                                  uint table_changes);
+  uint alter_table_flags(uint flags);
+#ifdef MRN_HANDLER_HAVE_ADD_INDEX
+  int add_index(TABLE *table_arg, KEY *key_info, uint num_of_keys,
+                handler_add_index **add);
+  int final_add_index(handler_add_index *add, bool commit);
+  int prepare_drop_index(TABLE *table_arg, uint *key_num, uint num_of_keys);
+  int final_drop_index(TABLE *table_arg);
+#endif
+  int update_auto_increment();
+  void set_next_insert_id(ulonglong id);
+  void get_auto_increment(ulonglong offset, ulonglong increment, ulonglong nb_desired_values,
+                          ulonglong *first_value, ulonglong *nb_reserved_values);
+  void restore_auto_increment(ulonglong prev_insert_id);
+  void release_auto_increment();
+  int reset_auto_increment(ulonglong value);
 
 protected:
 #ifdef MRN_HANDLER_HAVE_HA_RND_NEXT
@@ -292,12 +326,21 @@ protected:
 #endif
 
 private:
+  void push_warning_unsupported_spatial_index_search(enum ha_rkey_function flag);
+  void clear_cursor();
+  void clear_search_result();
+  grn_obj *find_tokenizer(const char *name, int name_length);
+  int storage_get_next_record(uchar *buf);
+  grn_obj *storage_geo_select_in_rectangle(grn_obj *index_column,
+                                           const uchar *rectangle);
+
 #ifdef MRN_HANDLER_HAVE_HA_CLOSE
   int close();
 #endif
   void check_count_skip(key_part_map start_key_part_map,
                         key_part_map end_key_part_map, bool fulltext);
-  void check_fast_order_limit();
+  void check_fast_order_limit(grn_table_sort_key **sort_keys, int *n_sort_keys,
+                              longlong *limit, grn_obj *score_column);
   void store_fields_from_primary_table(uchar *buf, grn_id record_id);
   void set_pk_bitmap();
   int wrapper_create(const char *name, TABLE *table,
@@ -309,11 +352,14 @@ private:
                                  const char *grn_table_name,
                                  int i,
                                  KEY *key_info,
-                                 grn_obj **index_tables);
+                                 grn_obj **index_tables,
+                                 MRN_SHARE *tmp_share);
   int wrapper_create_index(const char *name, TABLE *table,
                            HA_CREATE_INFO *info, MRN_SHARE *tmp_share);
   int storage_create_validate_pseudo_column(TABLE *table);
   int storage_create_validate_index(TABLE *table);
+  int storage_create_index(TABLE *table, const char *grn_table_name,
+                           grn_obj *grn_table, MRN_SHARE *tmp_share);
   int ensure_database_create(const char *name);
   int ensure_database_open(const char *name);
   int wrapper_delete_table(const char *name, MRN_SHARE *tmp_share,
@@ -353,13 +399,16 @@ private:
   int wrapper_write_row(uchar *buf);
   int wrapper_write_row_index(uchar *buf);
   int storage_write_row(uchar *buf);
+  int storage_write_row_index(uchar *buf, grn_id record_id);
   int wrapper_get_record_id(uchar *data, grn_id *record_id, const char *context);
   int wrapper_update_row(const uchar *old_data, uchar *new_data);
   int wrapper_update_row_index(const uchar *old_data, uchar *new_data);
   int storage_update_row(const uchar *old_data, uchar *new_data);
+  int storage_update_row_index(const uchar *old_data, uchar *new_data);
   int wrapper_delete_row(const uchar *buf);
   int wrapper_delete_row_index(const uchar *buf);
   int storage_delete_row(const uchar *buf);
+  int storage_delete_row_index(const uchar *buf);
   uint wrapper_max_supported_key_parts();
   uint storage_max_supported_key_parts();
   ulonglong wrapper_table_flags() const;
@@ -382,16 +431,21 @@ private:
                                    key_range *range_max);
   ha_rows storage_records_in_range(uint key_nr, key_range *range_min,
                                    key_range *range_max);
+  ha_rows storage_records_in_range_geo(uint key_nr, key_range *range_min,
+                                       key_range *range_max);
   int wrapper_index_init(uint idx, bool sorted);
   int storage_index_init(uint idx, bool sorted);
   int wrapper_index_end();
   int storage_index_end();
-  int wrapper_index_read_map(uchar * buf, const uchar * key,
+  int wrapper_index_read_map(uchar *buf, const uchar *key,
                              key_part_map keypart_map,
                              enum ha_rkey_function find_flag);
-  int storage_index_read_map(uchar * buf, const uchar * key,
+  int storage_index_read_map(uchar *buf, const uchar *key,
                              key_part_map keypart_map,
                              enum ha_rkey_function find_flag);
+  int storage_index_read_map_geo(uchar *buf, const uchar *key,
+                                 Field *field,
+                                 enum ha_rkey_function find_flag);
   int wrapper_index_read_last_map(uchar *buf, const uchar *key,
                                   key_part_map keypart_map);
   int storage_index_read_last_map(uchar *buf, const uchar *key,
@@ -414,6 +468,7 @@ private:
                                bool eq_range, bool sorted);
   int wrapper_read_range_next();
   int storage_read_range_next();
+  int generic_ft_init();
   int wrapper_ft_init();
   int storage_ft_init();
   void merge_matched_record_keys(grn_obj *matched_result);
@@ -421,6 +476,7 @@ private:
   void storage_ft_end();
   FT_INFO *wrapper_ft_init_ext(uint flags, uint key_nr, String *key);
   FT_INFO *storage_ft_init_ext(uint flags, uint key_nr, String *key);
+  FT_INFO *generic_ft_init_ext(uint flags, uint key_nr, String *key);
   int wrapper_ft_read(uchar *buf);
   int storage_ft_read(uchar *buf);
   const Item *wrapper_cond_push(const Item *cond);
@@ -519,6 +575,44 @@ private:
   int storage_optimize(THD* thd, HA_CHECK_OPT* check_opt);
   bool wrapper_is_fatal_error(int error_num, uint flags);
   bool storage_is_fatal_error(int error_num, uint flags);
+  bool wrapper_check_if_incompatible_data(HA_CREATE_INFO *create_info,
+                                          uint table_changes);
+  bool storage_check_if_incompatible_data(HA_CREATE_INFO *create_info,
+                                          uint table_changes);
+  uint wrapper_alter_table_flags(uint flags);
+  uint storage_alter_table_flags(uint flags);
+#ifdef MRN_HANDLER_HAVE_ADD_INDEX
+  int wrapper_add_index(TABLE *table_arg, KEY *key_info, uint num_of_keys,
+                        handler_add_index **add);
+  int storage_add_index(TABLE *table_arg, KEY *key_info, uint num_of_keys,
+                        handler_add_index **add);
+  int wrapper_final_add_index(handler_add_index *add, bool commit);
+  int storage_final_add_index(handler_add_index *add, bool commit);
+  int wrapper_prepare_drop_index(TABLE *table_arg, uint *key_num,
+                                 uint num_of_keys);
+  int storage_prepare_drop_index(TABLE *table_arg, uint *key_num,
+                                 uint num_of_keys);
+  int wrapper_final_drop_index(TABLE *table_arg);
+  int storage_final_drop_index(TABLE *table_arg);
+#endif
+  int wrapper_update_auto_increment();
+  int storage_update_auto_increment();
+  void wrapper_set_next_insert_id(ulonglong id);
+  void storage_set_next_insert_id(ulonglong id);
+  void wrapper_get_auto_increment(ulonglong offset, ulonglong increment,
+                                  ulonglong nb_desired_values,
+                                  ulonglong *first_value,
+                                  ulonglong *nb_reserved_values);
+  void storage_get_auto_increment(ulonglong offset, ulonglong increment,
+                                  ulonglong nb_desired_values,
+                                  ulonglong *first_value,
+                                  ulonglong *nb_reserved_values);
+  void wrapper_restore_auto_increment(ulonglong prev_insert_id);
+  void storage_restore_auto_increment(ulonglong prev_insert_id);
+  void wrapper_release_auto_increment();
+  void storage_release_auto_increment();
+  int wrapper_reset_auto_increment(ulonglong value);
+  int storage_reset_auto_increment(ulonglong value);
 };
 
 #ifdef __cplusplus
